@@ -1,5 +1,5 @@
 /* GMM (C) 2018 Stephane Charette <stephanecharette@gmail.com>
- * $Id: TabCanvas.cpp 2659 2018-11-11 08:54:01Z stephane $
+ * $Id: TabCanvas.cpp 2683 2018-12-05 08:44:20Z stephane $
  */
 
 #include "GMM.hpp"
@@ -8,16 +8,28 @@
 TabCanvas::TabCanvas(const std::string & name) :
 	Thread("thread for device prefix " + name),
 	prefix(name),
+	opc_sequence_number(0),
+	opc_output_filename(cfg().get_str("opc_output_filename")),
+	opc_register_kiln(cfg().get_str(prefix + "_opc_kiln_register") + "="),
+	opc_register_run(cfg().get_str(prefix + "_opc_run_register") + "="),
+	opc_register_pkg(cfg().get_str(prefix + "_opc_pkg_register") + "="),
 	tab_is_active(false),
 	export_image(cfg().get_bool(prefix + "_export_image")),
 	timer_frequency_in_milliseconds(cfg().get_int(prefix + "_chart_redraw_ms")),
 	comm_handle(INVALID_HANDLE_VALUE),
+	current_kiln(0),
+	current_run(0),
+	current_pkg(0),
+	new_kiln(cfg().get_int(prefix + "_last_kiln")),
+	new_run(cfg().get_int(prefix + "_last_run")),
+	new_pkg(cfg().get_int(prefix + "_last_pkg")),
 	run_sum(0.0),
 	pkg_sum(0.0),
 	run_average(0.0),
 	pkg_average(0.0),
 	run_std_dev(0.0),
-	pkg_std_dev(0.0)
+	pkg_std_dev(0.0),
+	opc_via_csv_filename(cfg().get_str(prefix + "_opc_via_csv_filename"))
 {
 	const Colour pink_background(0xe6, 0xcc, 0xf2);	// total
 	const Colour blue_background(0xc0, 0xc0, 0xff);	// local
@@ -32,18 +44,18 @@ TabCanvas::TabCanvas(const std::string & name) :
 	addAndMakeVisible(run_editor					);
 	addAndMakeVisible(pkg_label						);
 	addAndMakeVisible(pkg_editor					);
-	addAndMakeVisible(total_number_of_boards_label	);
-	addAndMakeVisible(total_number_of_boards_editor	);
-	addAndMakeVisible(total_average_moisture_label	);
-	addAndMakeVisible(total_average_moisture_editor	);
-	addAndMakeVisible(total_std_dev_moisture_label	);
-	addAndMakeVisible(total_std_dev_moisture_editor	);
-	addAndMakeVisible(local_number_of_boards_label	);
-	addAndMakeVisible(local_number_of_boards_editor	);
-	addAndMakeVisible(local_average_moisture_label	);
-	addAndMakeVisible(local_average_moisture_editor	);
-	addAndMakeVisible(local_std_dev_moisture_label	);
-	addAndMakeVisible(local_std_dev_moisture_editor	);
+	addAndMakeVisible(run_number_of_boards_label	);
+	addAndMakeVisible(run_number_of_boards_editor	);
+	addAndMakeVisible(run_average_moisture_label	);
+	addAndMakeVisible(run_average_moisture_editor	);
+	addAndMakeVisible(run_std_dev_moisture_label	);
+	addAndMakeVisible(run_std_dev_moisture_editor	);
+	addAndMakeVisible(pkg_number_of_boards_label	);
+	addAndMakeVisible(pkg_number_of_boards_editor	);
+	addAndMakeVisible(pkg_average_moisture_label	);
+	addAndMakeVisible(pkg_average_moisture_editor	);
+	addAndMakeVisible(pkg_std_dev_moisture_label	);
+	addAndMakeVisible(pkg_std_dev_moisture_editor	);
 
 	kiln_num_label	.setText("Kiln:", NotificationType::sendNotification);
 	run_label		.setText("Run:"	, NotificationType::sendNotification);
@@ -56,39 +68,39 @@ TabCanvas::TabCanvas(const std::string & name) :
 	run_editor		.setReadOnly(true);		run_editor		.setJustification(Justification::centredRight);
 	pkg_editor		.setReadOnly(true);		pkg_editor		.setJustification(Justification::centredRight);
 
-	total_number_of_boards_label	.setText("Boards:"	, NotificationType::sendNotification);
-	total_average_moisture_label	.setText("Avg:"		, NotificationType::sendNotification);
-	total_std_dev_moisture_label	.setText("StDv:"	, NotificationType::sendNotification);
-	total_number_of_boards_label	.setJustificationType(Justification::centredRight);
-	total_average_moisture_label	.setJustificationType(Justification::centredRight);
-	total_std_dev_moisture_label	.setJustificationType(Justification::centredRight);
+	run_number_of_boards_label	.setText("Boards:"	, NotificationType::sendNotification);
+	run_average_moisture_label	.setText("Avg:"		, NotificationType::sendNotification);
+	run_std_dev_moisture_label	.setText("StDv:"	, NotificationType::sendNotification);
+	run_number_of_boards_label	.setJustificationType(Justification::centredRight);
+	run_average_moisture_label	.setJustificationType(Justification::centredRight);
+	run_std_dev_moisture_label	.setJustificationType(Justification::centredRight);
 
-	total_number_of_boards_editor	.setReadOnly(true);
-	total_average_moisture_editor	.setReadOnly(true);
-	total_std_dev_moisture_editor	.setReadOnly(true);
-	total_number_of_boards_editor	.setJustification(Justification::centredRight);
-	total_average_moisture_editor	.setJustification(Justification::centredRight);
-	total_std_dev_moisture_editor	.setJustification(Justification::centredRight);
-	total_number_of_boards_editor	.setColour(TextEditor::backgroundColourId, pink_background);
-	total_average_moisture_editor	.setColour(TextEditor::backgroundColourId, pink_background);
-	total_std_dev_moisture_editor	.setColour(TextEditor::backgroundColourId, pink_background);
+	run_number_of_boards_editor	.setReadOnly(true);
+	run_average_moisture_editor	.setReadOnly(true);
+	run_std_dev_moisture_editor	.setReadOnly(true);
+	run_number_of_boards_editor	.setJustification(Justification::centredRight);
+	run_average_moisture_editor	.setJustification(Justification::centredRight);
+	run_std_dev_moisture_editor	.setJustification(Justification::centredRight);
+	run_number_of_boards_editor	.setColour(TextEditor::backgroundColourId, pink_background);
+	run_average_moisture_editor	.setColour(TextEditor::backgroundColourId, pink_background);
+	run_std_dev_moisture_editor	.setColour(TextEditor::backgroundColourId, pink_background);
 
-	local_number_of_boards_label	.setText("Boards:"	, NotificationType::sendNotification);
-	local_average_moisture_label	.setText("Avg:"		, NotificationType::sendNotification);
-	local_std_dev_moisture_label	.setText("StDv:"	, NotificationType::sendNotification);
-	local_number_of_boards_label	.setJustificationType(Justification::centredRight);
-	local_average_moisture_label	.setJustificationType(Justification::centredRight);
-	local_std_dev_moisture_label	.setJustificationType(Justification::centredRight);
+	pkg_number_of_boards_label	.setText("Boards:"	, NotificationType::sendNotification);
+	pkg_average_moisture_label	.setText("Avg:"		, NotificationType::sendNotification);
+	pkg_std_dev_moisture_label	.setText("StDv:"	, NotificationType::sendNotification);
+	pkg_number_of_boards_label	.setJustificationType(Justification::centredRight);
+	pkg_average_moisture_label	.setJustificationType(Justification::centredRight);
+	pkg_std_dev_moisture_label	.setJustificationType(Justification::centredRight);
 
-	local_number_of_boards_editor	.setReadOnly(true);
-	local_average_moisture_editor	.setReadOnly(true);
-	local_std_dev_moisture_editor	.setReadOnly(true);
-	local_number_of_boards_editor	.setJustification(Justification::centredRight);
-	local_average_moisture_editor	.setJustification(Justification::centredRight);
-	local_std_dev_moisture_editor	.setJustification(Justification::centredRight);
-	local_number_of_boards_editor	.setColour(TextEditor::backgroundColourId, blue_background);
-	local_average_moisture_editor	.setColour(TextEditor::backgroundColourId, blue_background);
-	local_std_dev_moisture_editor	.setColour(TextEditor::backgroundColourId, blue_background);
+	pkg_number_of_boards_editor	.setReadOnly(true);
+	pkg_average_moisture_editor	.setReadOnly(true);
+	pkg_std_dev_moisture_editor	.setReadOnly(true);
+	pkg_number_of_boards_editor	.setJustification(Justification::centredRight);
+	pkg_average_moisture_editor	.setJustification(Justification::centredRight);
+	pkg_std_dev_moisture_editor	.setJustification(Justification::centredRight);
+	pkg_number_of_boards_editor	.setColour(TextEditor::backgroundColourId, blue_background);
+	pkg_average_moisture_editor	.setColour(TextEditor::backgroundColourId, blue_background);
+	pkg_std_dev_moisture_editor	.setColour(TextEditor::backgroundColourId, blue_background);
 
 	// ****************************
 	// left of the moisture buckets
@@ -109,27 +121,30 @@ TabCanvas::TabCanvas(const std::string & name) :
 		run_moisture_bucket[idx] = 0;
 		pkg_moisture_bucket[idx] = 0;
 
-		addAndMakeVisible(total_distribution[idx]);
+		addAndMakeVisible(run_distribution	[idx]);
 		addAndMakeVisible(percentages		[idx]);
-		addAndMakeVisible(local_distribution[idx]);
+		addAndMakeVisible(pkg_distribution	[idx]);
 
-		total_distribution[idx].setColour(TextEditor::backgroundColourId, pink_background);
-		local_distribution[idx].setColour(TextEditor::backgroundColourId, blue_background);
+		run_distribution[idx].setColour(TextEditor::backgroundColourId, pink_background);
+		pkg_distribution[idx].setColour(TextEditor::backgroundColourId, blue_background);
 
-		total_distribution[idx].setReadOnly(true);
-		local_distribution[idx].setReadOnly(true);
+		run_distribution[idx].setReadOnly(true);
+		pkg_distribution[idx].setReadOnly(true);
 
-		total_distribution[idx].setJustification(Justification::centredRight);
-		local_distribution[idx].setJustification(Justification::centredRight);
+		run_distribution[idx].setJustification(Justification::centredRight);
+		pkg_distribution[idx].setJustification(Justification::centredRight);
 
 		percentages[idx].setText(String(percentage_of_first_bucket + idx) + "%", NotificationType::sendNotification);
 		percentages[idx].setJustificationType(Justification::centredRight);
 	}
 
-	// insert a few fake numbers
-	kiln_num_editor	.setText("7");
-	run_editor		.setText("026");
-	pkg_editor		.setText("118");
+	if (opc_via_csv_filename.empty() == false)
+	{
+		opc_via_csv_regex = std::regex(cfg().get_str(prefix + "_opc_via_csv_regex"));
+	}
+
+	// re-populate the kiln, run, and package numbers
+	update_kiln_run_pkg();
 
 	read_all_mc_files();
 
@@ -164,7 +179,7 @@ void TabCanvas::timerCallback(void)
 	/* This is called periodically using a timer.  Usually at 1-second intervals, though this can be tweaked via
 	 * configuration.
 	 *
-	 * Note that since we stop the timer, do some redrawing and calculations, and then restart the drawing, it means
+	 * Note that since we stop the timer, do some redrawing and calculations, and then restart the timer, it means
 	 * the timer will significantly drift over time.  This should not be a concern since all we're doing is updating
 	 * portions of the screen, including the chart.
 	 */
@@ -175,6 +190,14 @@ void TabCanvas::timerCallback(void)
 	{
 		periodic_update();
 
+		// There are 2 different ways we can update the kiln/run/package numbers.  The first is with the OPC Console,
+		// and the second is with the automatic .csv file that Curt wrote in VB.  The .csv solution was temporary and
+		// the OPC Console method is the prefered method.  The code still attempts both, and returns if the .csv or
+		// OPC Console output file does not exist on disk.
+
+		parse_opc_console_output();
+		read_opc_values_from_csv();
+
 		if (timer_frequency_in_milliseconds > 0)
 		{
 			startTimer(timer_frequency_in_milliseconds); // in milliseconds
@@ -195,18 +218,18 @@ void TabCanvas::periodic_update(void)
 	// re-populate the fields with the amounts stored in the buckets
 	for (size_t idx=0; idx < number_of_buckets; idx++)
 	{
-		total_distribution[idx].setText(String(run_moisture_bucket[idx]));
-		local_distribution[idx].setText(String(pkg_moisture_bucket[idx]));
+		run_distribution[idx].setText(String(run_moisture_bucket[idx]));
+		pkg_distribution[idx].setText(String(pkg_moisture_bucket[idx]));
 	}
 
 	if (threadShouldExit() == false)
 	{
-		total_number_of_boards_editor.setText(String(run_doubles.size()));
-		local_number_of_boards_editor.setText(String(pkg_doubles.size()));
-		total_average_moisture_editor.setText(Lox::Numbers::format(run_average, 2));
-		local_average_moisture_editor.setText(Lox::Numbers::format(pkg_average, 2));
-		total_std_dev_moisture_editor.setText(Lox::Numbers::format(run_std_dev, 2));
-		local_std_dev_moisture_editor.setText(Lox::Numbers::format(pkg_std_dev, 2));
+		run_number_of_boards_editor.setText(String(run_doubles.size()));
+		pkg_number_of_boards_editor.setText(String(pkg_doubles.size()));
+		run_average_moisture_editor.setText(Lox::Numbers::format(run_average, 2));
+		pkg_average_moisture_editor.setText(Lox::Numbers::format(pkg_average, 2));
+		run_std_dev_moisture_editor.setText(Lox::Numbers::format(run_std_dev, 2));
+		pkg_std_dev_moisture_editor.setText(Lox::Numbers::format(pkg_std_dev, 2));
 	}
 
 //	LOG_MSG("prefix=" << prefix << ": periodic update for RUN: number_of_boards: " << run_doubles.size() << ", average: " << run_average << ", stddev: " << run_std_dev);
@@ -262,6 +285,8 @@ void TabCanvas::calculate_averages_and_std_deviations()
 
 	if (threadShouldExit() == false)
 	{
+		const ScopedLock scope_lock(numbers_and_calculations_mutex);
+
 		run_std_dev = calc_std_dev(run_doubles, run_average);
 		pkg_std_dev = calc_std_dev(pkg_doubles, pkg_average);
 	}
@@ -330,6 +355,8 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 		g.drawRect(x * width_float, 0.0f, 1.0f, height_float);
 	}
 
+	const ScopedLock scope_lock(numbers_and_calculations_mutex);
+
 	// determine the maximum value in all the buckets (initially set to "1" to prevent divide-by-zero later)
 	size_t max_run_bucket_value = 1;
 	size_t max_pkg_bucket_value = 1;
@@ -389,12 +416,14 @@ Image TabCanvas::draw_chart(const size_t width, const size_t height)
 		percentage ++;
 	}
 
-	// add warnings and errors to the top-left of the image
-	if (warnings_and_errors.empty() == false && threadShouldExit() == false)
+	// combine the two vectors of errors/warnings, and add them to the top-left of the image
+	auto tmp = gmm().wnd->warnings_and_errors;
+	tmp.insert(tmp.end(), warnings_and_errors.begin(), warnings_and_errors.end());
+	if (tmp.empty() == false && threadShouldExit() == false)
 	{
 		g.setColour(Colours::red);
 		float y = 0.0f;
-		for (const auto & str : warnings_and_errors)
+		for (const auto & str : tmp)
 		{
 			y += 20.0f;
 			g.drawText(str.c_str(), 10.0f, y, width_float - 20.0f, 20.0f, Justification::bottomLeft);
@@ -421,34 +450,34 @@ void TabCanvas::resized(void)
 	const float margin_size		= 5.0;
 	const FlexItem::Margin margin(margin_size);
 
-	FlexBox total_row_flexbox;
-	total_row_flexbox.flexDirection		= FlexBox::Direction::row;
-	total_row_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
-	total_row_flexbox.items.add(FlexItem(kiln_num_label		).withMargin(margin).withWidth(40));
-	total_row_flexbox.items.add(FlexItem(kiln_num_editor	).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(run_label			).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(run_editor			).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(pkg_label			).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(pkg_editor			).withMargin(margin).withFlex(1.0));
+	FlexBox run_row_flexbox;
+	run_row_flexbox.flexDirection		= FlexBox::Direction::row;
+	run_row_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
+	run_row_flexbox.items.add(FlexItem(kiln_num_label		).withMargin(margin).withWidth(40));
+	run_row_flexbox.items.add(FlexItem(kiln_num_editor	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_label			).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_editor			).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(pkg_label			).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(pkg_editor			).withMargin(margin).withFlex(1.0));
 	// insert a few blank spacers so the row lines up with the moisture distribution row below
 	for (size_t idx = 0; idx < number_of_buckets - 11; idx++)
 	{
-		total_row_flexbox.items.add(FlexItem().withMargin(margin).withFlex(1.0));
+		run_row_flexbox.items.add(FlexItem().withMargin(margin).withFlex(1.0));
 	}
-	total_row_flexbox.items.add(FlexItem(total_number_of_boards_label	).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(total_number_of_boards_editor	).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(total_average_moisture_label	).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(total_average_moisture_editor	).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(total_std_dev_moisture_label	).withMargin(margin).withFlex(1.0));
-	total_row_flexbox.items.add(FlexItem(total_std_dev_moisture_editor	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_number_of_boards_label	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_number_of_boards_editor	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_average_moisture_label	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_average_moisture_editor	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_std_dev_moisture_label	).withMargin(margin).withFlex(1.0));
+	run_row_flexbox.items.add(FlexItem(run_std_dev_moisture_editor	).withMargin(margin).withFlex(1.0));
 
-	FlexBox total_distribution_flexbox;
-	total_distribution_flexbox.flexDirection	= FlexBox::Direction::row;
-	total_distribution_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
-	total_distribution_flexbox.items.add(FlexItem(run_header).withMargin(margin).withWidth(40));
+	FlexBox run_distribution_flexbox;
+	run_distribution_flexbox.flexDirection	= FlexBox::Direction::row;
+	run_distribution_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
+	run_distribution_flexbox.items.add(FlexItem(run_header).withMargin(margin).withWidth(40));
 	for (size_t idx = 0; idx < number_of_buckets; idx++)
 	{
-		total_distribution_flexbox.items.add(FlexItem(total_distribution[idx]).withMargin(margin).withFlex(1.0));
+		run_distribution_flexbox.items.add(FlexItem(run_distribution[idx]).withMargin(margin).withFlex(1.0));
 	}
 
 	FlexBox percentage_flexbox;
@@ -460,30 +489,30 @@ void TabCanvas::resized(void)
 		percentage_flexbox.items.add(FlexItem(percentages[idx]).withMargin(margin).withFlex(1.0));
 	}
 
-	FlexBox local_distribution_flexbox;
-	local_distribution_flexbox.flexDirection	= FlexBox::Direction::row;
-	local_distribution_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
-	local_distribution_flexbox.items.add(FlexItem(pkg_header).withMargin(margin).withWidth(40));
+	FlexBox pkg_distribution_flexbox;
+	pkg_distribution_flexbox.flexDirection	= FlexBox::Direction::row;
+	pkg_distribution_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
+	pkg_distribution_flexbox.items.add(FlexItem(pkg_header).withMargin(margin).withWidth(40));
 	for (size_t idx=0; idx < number_of_buckets; idx++)
 	{
-		local_distribution_flexbox.items.add(FlexItem(local_distribution[idx]).withMargin(margin).withFlex(1.0));
+		pkg_distribution_flexbox.items.add(FlexItem(pkg_distribution[idx]).withMargin(margin).withFlex(1.0));
 	}
 
-	FlexBox local_row_flexbox;
-	local_row_flexbox.flexDirection		= FlexBox::Direction::row;
-	local_row_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
-	local_row_flexbox.items.add(FlexItem(/* spacer */).withMargin(margin).withWidth(40));
+	FlexBox pkg_row_flexbox;
+	pkg_row_flexbox.flexDirection		= FlexBox::Direction::row;
+	pkg_row_flexbox.justifyContent	= FlexBox::JustifyContent::spaceAround;
+	pkg_row_flexbox.items.add(FlexItem(/* spacer */).withMargin(margin).withWidth(40));
 	// insert a few blank spacers so the row lines up with the moisture distribution row below
 	for (size_t idx = 0; idx < number_of_buckets - 6; idx++)
 	{
-		local_row_flexbox.items.add(FlexItem().withMargin(margin).withFlex(1.0));
+		pkg_row_flexbox.items.add(FlexItem().withMargin(margin).withFlex(1.0));
 	}
-	local_row_flexbox.items.add(FlexItem(local_number_of_boards_label	).withMargin(margin).withFlex(1.0));
-	local_row_flexbox.items.add(FlexItem(local_number_of_boards_editor	).withMargin(margin).withFlex(1.0));
-	local_row_flexbox.items.add(FlexItem(local_average_moisture_label	).withMargin(margin).withFlex(1.0));
-	local_row_flexbox.items.add(FlexItem(local_average_moisture_editor	).withMargin(margin).withFlex(1.0));
-	local_row_flexbox.items.add(FlexItem(local_std_dev_moisture_label	).withMargin(margin).withFlex(1.0));
-	local_row_flexbox.items.add(FlexItem(local_std_dev_moisture_editor	).withMargin(margin).withFlex(1.0));
+	pkg_row_flexbox.items.add(FlexItem(pkg_number_of_boards_label	).withMargin(margin).withFlex(1.0));
+	pkg_row_flexbox.items.add(FlexItem(pkg_number_of_boards_editor	).withMargin(margin).withFlex(1.0));
+	pkg_row_flexbox.items.add(FlexItem(pkg_average_moisture_label	).withMargin(margin).withFlex(1.0));
+	pkg_row_flexbox.items.add(FlexItem(pkg_average_moisture_editor	).withMargin(margin).withFlex(1.0));
+	pkg_row_flexbox.items.add(FlexItem(pkg_std_dev_moisture_label	).withMargin(margin).withFlex(1.0));
+	pkg_row_flexbox.items.add(FlexItem(pkg_std_dev_moisture_editor	).withMargin(margin).withFlex(1.0));
 
 	FlexBox chart_row_flexbox;
 	chart_row_flexbox.flexDirection		= FlexBox::Direction::row;
@@ -494,11 +523,11 @@ void TabCanvas::resized(void)
 	// this is the "main" flexbox for the component
 	FlexBox flexbox;
 	flexbox.flexDirection									= FlexBox::Direction::column;
-	flexbox.items.add(FlexItem(total_row_flexbox			).withHeight(height));
-	flexbox.items.add(FlexItem(total_distribution_flexbox	).withHeight(height));
+	flexbox.items.add(FlexItem(run_row_flexbox				).withHeight(height));
+	flexbox.items.add(FlexItem(run_distribution_flexbox		).withHeight(height));
 	flexbox.items.add(FlexItem(percentage_flexbox			).withHeight(height));
-	flexbox.items.add(FlexItem(local_distribution_flexbox	).withHeight(height));
-	flexbox.items.add(FlexItem(local_row_flexbox			).withHeight(height));
+	flexbox.items.add(FlexItem(pkg_distribution_flexbox		).withHeight(height));
+	flexbox.items.add(FlexItem(pkg_row_flexbox				).withHeight(height));
 	flexbox.items.add(FlexItem(chart_row_flexbox).withFlex(2.0).withMargin(margin));
 
 	auto r = getLocalBounds().toFloat();
@@ -593,7 +622,7 @@ void TabCanvas::device_thread(void)
 				}
 				else
 				{
-					LOG_MSG(device_name << ": str=" << String(str).trim().toStdString() << ", reading=" << value);
+//					LOG_MSG(device_name << ": str=" << String(str).trim().toStdString() << ", reading=" << value);
 				}
 
 				if (value > 0.0)
@@ -603,18 +632,24 @@ void TabCanvas::device_thread(void)
 					if (idx < 0)					idx = 0;
 					if (idx >= number_of_buckets)	idx = number_of_buckets - 1;
 
-					run_doubles.push_back(value);
-					pkg_doubles.push_back(value);
+					if (true)
+					{
+						const ScopedLock scope_lock(numbers_and_calculations_mutex);
 
-					run_sum += value;
-					pkg_sum += value;
+						run_doubles.push_back(value);
+						pkg_doubles.push_back(value);
 
-					run_average = run_sum / static_cast<double>(run_doubles.size());
-					pkg_average = pkg_sum / static_cast<double>(pkg_doubles.size());
+						run_sum += value;
+						pkg_sum += value;
 
-					run_moisture_bucket[idx] ++;
-					pkg_moisture_bucket[idx] ++;
-					LOG_MSG(device_name << ": value=" << value << ", incrementing bucket #" << idx << " to " << pkg_moisture_bucket[idx]);
+						run_average = run_sum / static_cast<double>(run_doubles.size());
+						pkg_average = pkg_sum / static_cast<double>(pkg_doubles.size());
+
+						run_moisture_bucket[idx] ++;
+						pkg_moisture_bucket[idx] ++;
+					}
+
+//					LOG_MSG(device_name << ": value=" << value << ", incrementing bucket #" << idx << " to " << pkg_moisture_bucket[idx]);
 
 					// every once in a while, write what we have back out to disk so we can restore things when we come back up
 					if (0 == pkg_doubles.size() % 100)
@@ -661,6 +696,7 @@ void TabCanvas::device_thread(void)
 	LOG_MSG("prefix=" << prefix << ": detected shutdown signal");
 
 	write_mc_file();
+	write_summary_file();
 
 	LOG_MSG("prefix=" << prefix << ": device thread for \"" << tab_name << "\" on " << device_name << " is exiting");
 
@@ -678,7 +714,9 @@ void TabCanvas::write_mc_file(void)
 	File dir(dirname);
 	dir.createDirectory();
 
-	const std::string filename = dir.getChildFile(prefix + "_test.mc").getFullPathName().toStdString();
+	const ScopedLock scope_lock(numbers_and_calculations_mutex);
+
+	const std::string filename = dir.getChildFile(kiln_num_editor.getText() + "_" + run_editor.getText() + pkg_editor.getText() + ".mc").getFullPathName().toStdString();
 	LOG_MSG("prefix=" << prefix << ": writing data out to: " << filename << " (pkg size=" << pkg_doubles.size() << ", pkg avg=" <<  pkg_average << ")");
 
 	// LINE #1:
@@ -686,7 +724,7 @@ void TabCanvas::write_mc_file(void)
 	std::ofstream fs(filename);
 	fs << std::fixed << std::setw(4) << std::setfill('0') << pkg_doubles.size() << ",";
 
-	for (size_t idx = 0; idx < number_of_buckets - 6; idx++)
+	for (size_t idx = 0; idx < number_of_buckets; idx++)
 	{
 		fs << std::setw(4) << std::setfill('0') << pkg_moisture_bucket[idx] << ",";
 	}
@@ -709,6 +747,39 @@ void TabCanvas::write_mc_file(void)
 }
 
 
+void TabCanvas::write_summary_file(void)
+{
+	// make sure to see the Doxygen documentation to learn about all the fields
+
+	String dirname = cfg().get_str("output_directory");
+	dirname = dirname.replaceFirstOccurrenceOf("%Y", Time::getCurrentTime().formatted("%Y"));
+
+	File dir(dirname);
+	dir.createDirectory();
+
+	const ScopedLock scope_lock(numbers_and_calculations_mutex);
+
+	const std::string filename = dir.getChildFile(kiln_num_editor.getText() + "_" + run_editor.getText() + "000.mc").getFullPathName().toStdString();
+	LOG_MSG("prefix=" << prefix << ": writing summary data to: " << filename << " (pkg size=" << pkg_doubles.size() << ", pkg avg=" << pkg_average << ")");
+
+	std::ofstream fs(filename);
+	fs << run_doubles.size() << ",";
+
+	for (size_t idx = 0; idx < number_of_buckets; idx++)
+	{
+		fs << run_moisture_bucket[idx] << ",";
+	}
+
+	fs << "000,000,";	// unknown fields
+
+	fs << std::fixed << run_std_dev << "," << run_average << ",0.0,0.0" << std::endl;
+
+	LOG_MSG("prefix=" << prefix << ": done writing summary data to " << filename);
+
+	return;
+}
+
+
 void TabCanvas::read_all_mc_files(void)
 {
 	String dirname = cfg().get_str("output_directory");
@@ -717,52 +788,317 @@ void TabCanvas::read_all_mc_files(void)
 	File dir(dirname);
 	dir.createDirectory();
 
-	const std::string filename = dir.getChildFile(prefix + "_test.mc").getFullPathName().toStdString();
+	/* The .mc output files are named as follows:  K_RRRPPP.mc where:
+	 *
+	 *		K is the 1-digit kiln number.
+	 *		RRR is the 3-digit run number.
+	 *		PPP is the 3-digit package number.
+	 */
 
-	std::ifstream fs(filename);
-	std::string str;
-	std::getline(fs, str); // ignore the first line
-	std::getline(fs, str); // parse the 2nd line
-	fs.close();
-
-//	LOG_MSG("prefix=" << prefix << ": filename=" << filename << ": 2nd line to parse: \"" << str << "\"");
-
-	// replace comma with spaces to make it easier to parse the individual fields
-	while (true)
+	// note how index starts at 1, not zero, since index zero is the summary file
+	for (int package_index = 1; package_index <= 999; package_index ++)
 	{
-		const size_t pos = str.find(',');
-		if (pos == std::string::npos)
+		std::stringstream filename_ss;
+		filename_ss
+			<< current_kiln << "_"
+			<< std::setfill('0') << std::setw(3) << current_run
+			<< std::setfill('0') << std::setw(3) << package_index
+			<< ".mc";
+		File file = dir.getChildFile(filename_ss.str());
+
+		if (file.existsAsFile() == false)
 		{
-			break;
+			// this package doesn't exist, so look for the next one
+			continue;
+		}
+
+		// we're starting to read a new package, so clear out the package-specific values
+		for (size_t idx = 0; idx < number_of_buckets; idx++)
+		{
+			pkg_moisture_bucket[idx] = 0;
+		}
+		pkg_doubles.clear();
+		pkg_sum = 0.0;
+		pkg_std_dev = 0.0;
+
+		const std::string filename = file.getFullPathName().toStdString();
+		LOG_MSG("prefix=" << prefix << ": importing all existing values from " << filename);
+
+		std::ifstream fs(filename);
+		std::string str;
+		std::getline(fs, str); // ignore the first line
+		std::getline(fs, str); // parse the 2nd line
+		fs.close();
+
+		// replace comma with spaces to make it easier to parse the individual fields
+		while (true)
+		{
+			const size_t pos = str.find(',');
+			if (pos == std::string::npos)
+			{
+				break;
+			}
+			str[pos] = ' ';
+		}
+
+		const ScopedLock scope_lock(numbers_and_calculations_mutex);
+
+		size_t counter = 0;
+		std::stringstream ss(str);
+		while (ss)
+		{
+			ss >> str;
+			if (str.empty() == false && str != ".")
+			{
+				const double d = std::stod(str);
+				if (d > 0.0)
+				{
+					counter ++;
+					int idx = static_cast<int>(d) - percentage_of_first_bucket;
+					if (idx < 0)					idx = 0;
+					if (idx >= number_of_buckets)	idx = number_of_buckets - 1;
+
+					run_moisture_bucket[idx] ++;
+					pkg_moisture_bucket[idx] ++;
+
+					run_doubles.push_back(d);
+					pkg_doubles.push_back(d);
+
+					run_sum += d;
+					pkg_sum += d;
+				}
+			}
+		}
+
+		if (run_doubles.empty() == false)
+		{
+			run_average = run_sum / static_cast<double>(run_doubles.size());
+			LOG_MSG("prefix=" << prefix << ": RUN: filename=" << filename << ": imported " << counter << " existing entries, run average is " << run_average);
+		}
+		if (pkg_doubles.empty() == false)
+		{
+			pkg_average = pkg_sum / static_cast<double>(pkg_doubles.size());
+			LOG_MSG("prefix=" << prefix << ": PKG: filename=" << filename << ": imported " << counter << " existing entries, pkg average is " << pkg_average);
 		}
-		str[pos] = ' ';
 	}
 
-	std::stringstream ss(str);
-	while (ss)
-	{
-		ss >> str;
-		if (str.empty() == false && str != ".")
-		{
-			const double d = std::stod(str);
-			if (d > 0.0)
-			{
-				int idx = static_cast<int>(d) - percentage_of_first_bucket;
-				if (idx < 0)					idx = 0;
-				if (idx >= number_of_buckets)	idx = number_of_buckets - 1;
+	return;
+}
 
-				run_moisture_bucket[idx] ++;
-				run_doubles.push_back(d);
-				run_sum += d;
+
+void TabCanvas::parse_opc_console_output(void)
+{
+	//LOG_MSG("prefix=" << prefix << ": checking to see we need to parse OPC output file \"" << opc_output_filename << "\"");
+
+	const Lox::VStr v = Lox::readTextFile(opc_output_filename, false, false);
+	if (v.empty())
+	{
+		// not much we can do if we failed to read the file
+		LOG_MSG("prefix=" << prefix << ": failed to load the OPC output file \"" << opc_output_filename << "\"");
+		return;
+	}
+
+	/* The file should look something like this:
+	 *
+	 *	37
+	 *	[INFEED]N75:3=1
+	 *	[INFEED]N75:5=136
+	 *	[INFEED]N75:4=455
+	 *	[INFEED]N75:13=2
+	 *	[INFEED]N75:15=216
+	 *	[INFEED]N75:14=57
+	 *
+	 * The order of the lines is not guaranteed.
+	 */
+
+	// first line is the sequence number
+	const size_t new_sequence_number = std::stoul(v[0]);
+	if (new_sequence_number == opc_sequence_number)
+	{
+		// file hasn't changed, so nothing else we need to do
+		//LOG_MSG("prefix=" << prefix << ": OPC Console output file " << opc_output_filename << " has not changed (sequence number is \"" << opc_sequence_number << "\")");
+		return;
+	}
+	opc_sequence_number = new_sequence_number;
+	LOG_MSG("prefix=" << prefix << ": detected OPC Console output file " << opc_output_filename << " has a sequence number of \"" << opc_sequence_number << "\"");
+
+	// loop through the rest of the values to see if it has anything we need
+	for (size_t idx=1; idx < v.size(); idx++)
+	{
+		const std::string & line = v.at(idx);
+		size_t pos = line.find("=");
+		if (pos == std::string::npos)
+		{
+			// skip invalid lines
+			continue;
+		}
+
+		// Note we want (and need) the key to include the trailing "=" character,
+		// otherwise we could get some false positives since "N75:10" would match "N75:1",
+		// but "N75:10=" wont ever be a match for "N75:1=".
+		const std::string key = line.substr(0, pos + 1);
+		const int value = std::stoi(line.substr(pos + 1));
+
+//		LOG_MSG("prefix=" << prefix << ": -> idx=" << idx << ", key=" << key << " value=" << value);
+
+		if (key.find(opc_register_kiln)	!= std::string::npos)	new_kiln	= value;
+		if (key.find(opc_register_run)	!= std::string::npos)	new_run		= value;
+		if (key.find(opc_register_pkg)	!= std::string::npos)	new_pkg		= value;
+	}
+
+	update_kiln_run_pkg();
+
+	return;
+}
+
+
+void TabCanvas::read_opc_values_from_csv(void)
+{
+	if (opc_via_csv_filename.empty())
+	{
+		// nothing we can do if we don't know the .csv filename
+		return;
+	}
+
+	LOG_MSG("prefix=" << prefix << ": checking to see if there are any changes in \"" << opc_via_csv_filename << "\"");
+
+	const Lox::VStr v = Lox::readTextFile(opc_via_csv_filename, false, false);
+	for (const auto & line : v)
+	{
+		std::smatch matches;
+		const bool found = std::regex_search(line, matches, opc_via_csv_regex);
+		if (found)
+		{
+			const std::string key = matches[1];
+			const std::string val = matches[2];
+			const int value = std::stoi(val);
+
+			LOG_MSG("prefix=" << prefix << ": read from CSV: key=" << key << ", val=" << val);
+
+			if (key == "Kiln")
+			{
+				new_kiln = value;
+			}
+			else if (key == "Run")
+			{
+				new_run = value;
+			}
+			else if (key == "Pkg")
+			{
+				new_pkg = value;
 			}
 		}
 	}
 
-	if (run_doubles.empty() == false)
+	update_kiln_run_pkg();
+
+	return;
+}
+
+
+void TabCanvas::update_kiln_run_pkg(void)
+{
+	if (current_kiln == new_kiln	&&
+		current_run == new_run		&&
+		current_pkg == new_pkg		)
 	{
-		run_average = run_sum / static_cast<double>(run_doubles.size());
+		// nothing has changed
+		return;
 	}
-	LOG_MSG("prefix=" << prefix << ": filename=" << filename << ": imported " << run_doubles.size() << " existing entries, average is " << run_average);
+
+	bool need_to_output_mc_file = false;
+	bool run_needs_to_be_reset = false;
+	bool pkg_needs_to_be_reset = false;
+
+	if (current_kiln	> 0 &&
+		current_run		> 0 &&
+		current_pkg		> 0)
+	{
+		need_to_output_mc_file = true;
+
+		if (current_pkg != new_pkg)
+		{
+			pkg_needs_to_be_reset = true;
+		}
+		if (current_run != new_run || current_kiln != new_kiln)
+		{
+			// if the run changes, then both the run and the package both need to be reset
+			pkg_needs_to_be_reset = true;
+			run_needs_to_be_reset = true;
+		}
+	}
+
+	LOG_MSG("prefix=" << prefix << ": need to update the kiln, run, or package number");
+	LOG_MSG("prefix=" << prefix << ": -> current k=" << current_kiln	<< " r=" << current_run	<< " p=" << current_pkg	);
+	LOG_MSG("prefix=" << prefix << ": -> updated k=" << new_kiln		<< " r=" << new_run		<< " p=" << new_pkg		);
+	LOG_MSG("prefix=" << prefix << ": flag indicating we need to output .mc files " << (need_to_output_mc_file ? "has been set" : "is not set"));
+	LOG_MSG("prefix=" << prefix << ": flag indicating pkg variables will be reset: " << (pkg_needs_to_be_reset ? "true" : "false"));
+	LOG_MSG("prefix=" << prefix << ": flag indicating run variables will be reset: " << (run_needs_to_be_reset ? "true" : "false"));
+
+	if (need_to_output_mc_file)
+	{
+		const ScopedLock scope_lock(numbers_and_calculations_mutex);
+		write_mc_file();
+		write_summary_file();
+
+		for (size_t idx = 0; idx < number_of_buckets; idx++)
+		{
+			if (pkg_needs_to_be_reset)
+			{
+				pkg_moisture_bucket[idx] = 0;
+			}
+			if (run_needs_to_be_reset)
+			{
+				run_moisture_bucket[idx] = 0;
+			}
+		}
+
+		if (pkg_needs_to_be_reset)
+		{
+			pkg_doubles.clear();
+			pkg_sum		= 0.0;
+			pkg_average	= 0.0;
+			pkg_std_dev	= 0.0;
+		}
+		if (run_needs_to_be_reset)
+		{
+			run_doubles.clear();
+			run_sum		= 0.0;
+			run_average	= 0.0;
+			run_std_dev	= 0.0;
+		}
+	}
+
+	std::string kiln_str = std::to_string(new_kiln);
+	std::string run_str = std::to_string(new_run);
+	std::string pkg_str = std::to_string(new_pkg);
+
+	while (run_str.length() < 3)
+	{
+		run_str = "0" + run_str;
+	}
+	while (pkg_str.length() < 3)
+	{
+		pkg_str = "0" + pkg_str;
+	}
+
+	kiln_num_editor.setText(kiln_str);
+	run_editor.setText(run_str);
+	pkg_editor.setText(pkg_str);
+
+	current_kiln = new_kiln;
+	current_run = new_run;
+	current_pkg = new_pkg;
+
+	cfg().setValue(prefix + "_last_kiln", new_kiln);
+	cfg().setValue(prefix + "_last_run", new_run);
+	cfg().setValue(prefix + "_last_pkg", new_pkg);
+
+	// if the run was reset, re-read all the new files (if any) now that we've updated the run and package numbers
+	if (run_needs_to_be_reset)
+	{
+		read_all_mc_files();
+	}
 
 	return;
 }
