@@ -1,5 +1,5 @@
 /* GMM (C) 2018 Stephane Charette <stephanecharette@gmail.com>
- * $Id: TabCanvas.hpp 2677 2018-12-04 07:22:17Z stephane $
+ * $Id: TabCanvas.hpp 2682 2018-12-05 08:30:42Z stephane $
  */
 
 #pragma once
@@ -59,6 +59,8 @@ class TabCanvas : public Component, public Timer, public Thread
 
 		virtual void read_opc_values_from_csv(void);
 
+		virtual void parse_opc_console_output(void);
+
 		virtual void update_kiln_run_pkg(void);
 
 		/// Lock this mutex when modifying some of the key variables such as @ref run_doubles, @ref run_sum, @ref run_average, and @ref run_std_dev.
@@ -66,6 +68,12 @@ class TabCanvas : public Component, public Timer, public Thread
 
 		const std::string prefix;
 
+		size_t				opc_sequence_number;
+		const std::string	opc_output_filename;
+		const std::string	opc_register_kiln;
+		const std::string	opc_register_run;
+		const std::string	opc_register_pkg;
+
 		/** Used to determine if this tab is the currently-selected tab in the notebook.  Is updated by the Notebook
 		 * object whenever a new tab gets focus.  @see @ref Notebook::currentTabChanged()
 		 */
